package cc.fascinated.bat.features.base.commands.general;

import cc.fascinated.bat.Consts;
import cc.fascinated.bat.command.BatCommand;
import cc.fascinated.bat.command.Category;
import cc.fascinated.bat.command.CommandInfo;
import cc.fascinated.bat.common.EmbedUtils;
import cc.fascinated.bat.event.EventListener;
import cc.fascinated.bat.model.BatGuild;
import cc.fascinated.bat.model.BatUser;
import cc.fascinated.bat.service.CommandService;
import lombok.NonNull;
import net.dv8tion.jda.api.entities.Member;
import net.dv8tion.jda.api.entities.Message;
import net.dv8tion.jda.api.entities.MessageEmbed;
import net.dv8tion.jda.api.entities.MessageReference;
import net.dv8tion.jda.api.entities.channel.middleman.MessageChannel;
import net.dv8tion.jda.api.entities.emoji.Emoji;
import net.dv8tion.jda.api.events.interaction.component.StringSelectInteractionEvent;
import net.dv8tion.jda.api.interactions.commands.SlashCommandInteraction;
import net.dv8tion.jda.api.interactions.commands.build.SubcommandData;
import net.dv8tion.jda.api.interactions.components.ActionRow;
import net.dv8tion.jda.api.interactions.components.LayoutComponent;
import net.dv8tion.jda.api.interactions.components.buttons.Button;
import net.dv8tion.jda.api.interactions.components.buttons.ButtonStyle;
import net.dv8tion.jda.api.interactions.components.selections.SelectOption;
import net.dv8tion.jda.api.interactions.components.selections.StringSelectMenu;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * @author Fascinated (fascinated7)
 */
@Component
@CommandInfo(name = "help", description = "View the bots command categories.", guildOnly = false, userInstall = true, category = Category.GENERAL)
public class HelpCommand extends BatCommand implements EventListener {
    private final CommandService commandService;

    @Autowired
    public HelpCommand(@NonNull CommandService commandService) {
        this.commandService = commandService;
    }

    @Override
    public void execute(BatGuild guild, @NonNull BatUser user, @NonNull MessageChannel channel, Member member, @NonNull SlashCommandInteraction event) {
        event.replyEmbeds(createHomeEmbed()).addComponents(createHomeActions()).queue();
    }

    @Override
    public void onStringSelectInteraction(BatGuild guild, @NonNull BatUser user, @NonNull StringSelectInteractionEvent event) {
        if (event.getMember() == null) {
            return;
        }
        if (!event.getComponentId().startsWith("help")) {
            return;
        }

        // Get the item
        String item = event.getSelectedOptions().get(0).getValue();
        if (item.equalsIgnoreCase("home")) {
            event.editMessageEmbeds(createHomeEmbed()).queue();
            return;
        }

        // Get the category
        Category category = Category.getByName(item.split("-")[1]);
        if (category == null) {
            event.reply("Invalid category selected.")
                    .setEphemeral(true)
                    .queue();
            return;
        }

        // Send the category
        StringBuilder commands = new StringBuilder();
        List<BatCommand> categoryCommands = commandService.getCommandsByCategory(category);
        if (categoryCommands.isEmpty()) {
            commands = new StringBuilder("No commands available in this category.");
        } else {
            for (BatCommand command : categoryCommands) {
                if (!command.getSubCommands().isEmpty()) {
                    for (Map.Entry<String, BatCommand> entry : command.getSubCommands().entrySet()) {
                        BatCommand subCommand = entry.getValue();
                        SubcommandData subcommandData = subCommand.getSubcommandData();
                        commands.append("</%s %s:%s> - %s\n".formatted(
                                command.getInfo().getName(),
                                subcommandData.getName(),
                                subCommand.getSnowflake(),
                                subcommandData.getDescription()
                        ));
                    }
                    continue;
                }
                commands.append("</%s:%s> - %s\n".formatted(
                        command.getInfo().getName(),
                        command.getSnowflake(),
                        command.getInfo().getDescription()
                ));
            }
        }
        int subCommands = categoryCommands.stream().mapToInt(command -> command.getSubCommands().size()).sum();
        event.editMessageEmbeds(EmbedUtils.genericEmbed()
                .setAuthor("%s Category".formatted(category.getName()))
                .setDescription("%s command%s (with %s sub-command%s)\n\n**Commands:**\n%s".formatted(
                        categoryCommands.size(),
                        categoryCommands.size() == 1 ? "" : "s",
                        subCommands,
                        subCommands == 1 ? "" : "s",
                        commands.toString()
                )).build()).queue();
    }

    /**
     * Creates the home embed for the help command
     *
     * @return The home embed
     */
    private MessageEmbed createHomeEmbed() {
        StringBuilder categories = new StringBuilder();
        for (Category category : Category.values()) {
            long commandCount = commandService.getCommandsByCategory(category).size();
            categories.append("➜  %s - **%s Command%s**\n".formatted(
                    category.getName(),
                    commandCount,
                    commandCount == 1 ? "" : "s"
            ));
        }

        return EmbedUtils.genericEmbed()
                .setDescription("""
                        **Welcome to the Bat Help Menu!**
                        
                        %s
                        *View our [TOS](%s) and [Privacy Policy](%s) for more information.*
                        """.formatted(
                        categories.toString(),
                        Consts.TERMS_OF_SERVICE_URL,
                        Consts.PRIVACY_POLICY_URL
                ))
                .build();
    }

    /**
     * Creates the home actions for the help command
     *
     * @return The layout components
     */
    private LayoutComponent[] createHomeActions() {
        List<SelectOption> options = new ArrayList<>();
        options.add(SelectOption.of("Home", "help-home").withEmoji(Emoji.fromUnicode("U+1F3E0")));
        for (Category category : Category.values()) {
            options.add(SelectOption.of(category.getName(), "help-" + category.getName()).withEmoji(category.getEmoji()));
        }

        return new LayoutComponent[]{
                ActionRow.of(
                        Button.of(ButtonStyle.LINK, Consts.INVITE_URL, "Invite"),
                        Button.of(ButtonStyle.LINK, Consts.SUPPORT_INVITE_URL, "Support")
                ),
                ActionRow.of(
                        StringSelectMenu.create("help-menu")
                                .addOptions(options)
                                .build()
                )
        };
    }
}
