package cc.fascinated.bat.features.logging.listeners;

import cc.fascinated.bat.common.*;
import cc.fascinated.bat.event.EventListener;
import cc.fascinated.bat.features.logging.LogFeature;
import cc.fascinated.bat.features.logging.LogType;
import cc.fascinated.bat.model.BatGuild;
import cc.fascinated.bat.model.BatUser;
import cc.fascinated.bat.service.DiscordService;
import cc.fascinated.bat.service.UserService;
import jakarta.annotation.PostConstruct;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import net.dv8tion.jda.api.Permission;
import net.dv8tion.jda.api.Region;
import net.dv8tion.jda.api.entities.Member;
import net.dv8tion.jda.api.entities.channel.ChannelType;
import net.dv8tion.jda.api.entities.channel.concrete.TextChannel;
import net.dv8tion.jda.api.entities.channel.concrete.VoiceChannel;
import net.dv8tion.jda.api.entities.channel.unions.AudioChannelUnion;
import net.dv8tion.jda.api.entities.channel.unions.ChannelUnion;
import net.dv8tion.jda.api.events.channel.ChannelCreateEvent;
import net.dv8tion.jda.api.events.channel.ChannelDeleteEvent;
import net.dv8tion.jda.api.events.channel.update.*;
import net.dv8tion.jda.api.events.guild.override.GenericPermissionOverrideEvent;
import net.dv8tion.jda.api.events.guild.voice.GenericGuildVoiceEvent;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Component;

import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

/**
 * @author Fascinated (fascinated7)
 */
@Component
@Log4j2(topic = "Logging - Channel Listener")
public class ChannelListener implements EventListener {
    /**
     * A map of users and the last voice channel they were in
     */
    private final Map<BatUser, VoiceChannel> lastVoiceChannel = new HashMap<>();
    private final UserService userService;
    private final LogFeature logFeature;

    @Autowired
    public ChannelListener(@NonNull ApplicationContext context, @NonNull UserService userService) {
        this.userService = userService;
        this.logFeature = context.getBean(LogFeature.class);
    }

    @PostConstruct
    public void init() {
        log.info("Getting last voice channel for all users");
        for (VoiceChannel voiceChannel : DiscordService.JDA.getVoiceChannels()) {
            for (Member member : voiceChannel.getMembers()) {
                BatUser user = userService.getUser(member.getId());
                if (user == null) {
                    continue;
                }
                lastVoiceChannel.put(user, voiceChannel);
            }
        }
        log.info("Found {} users in a voice channel", lastVoiceChannel.size());
    }

    /**
     * Formats the channel type
     *
     * @param channel - the channel
     * @return the formatted channel type
     */
    public String formatChannelType(ChannelUnion channel) {
        return (channel.getType() == ChannelType.TEXT || channel.getType() == ChannelType.VOICE) ? EnumUtils.getEnumName(channel.getType()) + " Channel"
                : EnumUtils.getEnumName(channel.getType());
    }

    @Override
    public void onChannelCreate(@NonNull BatGuild guild, @NonNull ChannelCreateEvent event) {
        ChannelUnion channel = event.getChannel();
        String type = formatChannelType(channel);
        log.info("{} \"{}\" was created in guild \"{}\"", type, event.getChannel().getName(), guild.getName());
        logFeature.sendLog(guild, LogType.CHANNEL_CREATE, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("%s Created".formatted(type))
                        .appendLine("%s: %s".formatted(type, event.getChannel().getAsMention()), true)
                        .appendLine("Name: `%s`".formatted(event.getChannel().getName()), true)
                        .build())
                .build());
    }

    @Override
    public void onChannelDelete(@NonNull BatGuild guild, @NonNull ChannelDeleteEvent event) {
        ChannelUnion channel = event.getChannel();
        String type = formatChannelType(channel);
        log.info("{} \"{}\" was deleted in guild \"{}\"", type, event.getChannel().getName(), guild.getName());
        EmbedDescriptionBuilder description = new EmbedDescriptionBuilder("%s Deleted".formatted(type))
                .appendLine("Name: `#%s`".formatted(channel.getName()), true);
        if (channel instanceof TextChannel) {
            TextChannel textChannel = channel.asTextChannel();
            description.appendLine("Topic: `%s`".formatted(textChannel.getTopic() == null ? "No Topic" : textChannel.getTopic()), true);
        }
        logFeature.sendLog(guild, LogType.CHANNEL_DELETE, EmbedUtils.errorEmbed().setDescription(description.build()).build());
    }

    @Override
    public void onGuildVoiceUpdate(@NonNull BatGuild guild, @NonNull BatUser user, @NonNull GenericGuildVoiceEvent event) {
        AudioChannelUnion channel = event.getVoiceState().getChannel();
        if (channel != null) {
            VoiceChannel voiceChannel = channel.asVoiceChannel();
            lastVoiceChannel.put(user, voiceChannel);
        }
        VoiceChannel voiceChannel = lastVoiceChannel.get(user);
        if (voiceChannel == null) {
            return;
        }
        boolean joined = voiceChannel.getMembers().contains(event.getMember());
        if (!joined) {
            lastVoiceChannel.remove(user);
        }
        log.info("User \"{}\" {} voice channel \"{}\" in guild \"{}\"", user.getName(), joined ? "joined" : "left", voiceChannel.getName(), guild.getName());
        String description = new EmbedDescriptionBuilder("Member %s Voice Channel".formatted(joined ? "Joined" : "Left"))
                .appendLine("User: %s".formatted(user.getDiscordUser().getAsMention()), true)
                .appendLine("Channel: %s".formatted(voiceChannel.getAsMention()), true)
                .build();
        if (joined) {
            logFeature.sendLog(guild, LogType.VOICE_CHANNEL_JOIN, EmbedUtils.successEmbed().setDescription(description).build());
            return;
        }
        logFeature.sendLog(guild, LogType.VOICE_CHANNEL_LEAVE, EmbedUtils.errorEmbed().setDescription(description).build());
    }

    @Override
    public void onPermissionOverride(@NonNull BatGuild guild, @NonNull GenericPermissionOverrideEvent event) {
        log.info("Channel permissions updated for channel \"{}\" in guild \"{}\"", event.getChannel().getName(), guild.getName());

        StringBuilder allowedPermissions = new StringBuilder();
        for (Permission permission : event.getPermissionOverride().getAllowed()) {
            allowedPermissions.append("`").append(permission.getName()).append("`, ");
        }

        StringBuilder deniedPermissions = new StringBuilder();
        for (Permission permission : event.getPermissionOverride().getDenied()) {
            deniedPermissions.append("`").append(permission.getName()).append("`, ");
        }

        String target = event.getPermissionOverride().isRoleOverride() ? event.getPermissionOverride().getRole().getAsMention()
                : event.getPermissionOverride().getMember().getAsMention();
        EmbedDescriptionBuilder description = new EmbedDescriptionBuilder("Channel permissions updated")
                .appendLine("Channel: %s".formatted(event.getChannel().getAsMention()), true)
                .appendLine("%s: %s".formatted(
                        event.getPermissionOverride().isRoleOverride() ? "Role" : "Member",
                        target
                ), true);

        if (!allowedPermissions.isEmpty()) {
            description.appendLine("Allowed Permissions: %s".formatted(allowedPermissions.toString().isEmpty() ? "None"
                    : allowedPermissions.substring(0, allowedPermissions.length() - 2)), true);
        }
        if (!deniedPermissions.isEmpty()) {
            description.appendLine("Denied Permissions: %s".formatted(deniedPermissions.toString().isEmpty() ? "None"
                    : deniedPermissions.substring(0, deniedPermissions.length() - 2)), true);
        }

        logFeature.sendLog(guild, LogType.CHANNEL_PERMISSIONS_UPDATED, EmbedUtils.successEmbed()
                .setDescription(description.build()).build());
    }

    @Override
    public void onChannelUpdateTopic(@NonNull BatGuild guild, @NonNull ChannelUpdateTopicEvent event) {
        String oldValue = event.getOldValue();
        String newValue = event.getNewValue();
        ChannelUnion channel = event.getChannel();
        String type = formatChannelType(channel);

        log.info("{} \"{}\" topic was updated to {} in guild \"{}\"", type, newValue, oldValue, guild.getName());
        EmbedDescriptionBuilder description = new EmbedDescriptionBuilder("%s Topic Updated".formatted(type));
        description.appendLine("Channel: %s".formatted(event.getChannel().getAsMention()), true);
        description.appendLine("Topic: `%s` -> `%s`".formatted(
                oldValue == null ? "No Topic" : oldValue,
                newValue == null ? "No Topic" : newValue
        ), true);
        logFeature.sendLog(guild, LogType.CHANNEL_CONFIGURATION, EmbedUtils.successEmbed()
                .setDescription(description.build())
                .build());
    }

    @Override
    public void onChannelUpdateBitrate(@NonNull BatGuild guild, @NonNull ChannelUpdateBitrateEvent event) {
        if (Objects.equals(event.getOldValue(), event.getNewValue()) || event.getOldValue() == null || event.getNewValue() == null) {
            return;
        }

        VoiceChannel channel = event.getChannel().asVoiceChannel();
        log.info("Voice channel \"{}\" bitrate was updated to {} in guild \"{}\"", channel.getName(), event.getNewValue(), guild.getName());
        EmbedDescriptionBuilder description = new EmbedDescriptionBuilder("Voice Channel Updated")
                .appendLine("Channel: %s".formatted(channel.getAsMention()), true)
                .appendLine("Bitrate: `%s` -> `%s`".formatted(NumberFormatter.format(event.getOldValue()),
                        NumberFormatter.format(event.getNewValue())), true);
        logFeature.sendLog(guild, LogType.CHANNEL_CONFIGURATION, EmbedUtils.successEmbed()
                .setDescription(description.build())
                .build());
    }

    @Override
    public void onChannelUpdateName(@NonNull BatGuild guild, @NonNull ChannelUpdateNameEvent event) {
        String oldValue = event.getOldValue();
        String newValue = event.getNewValue();
        ChannelUnion channel = event.getChannel();
        String type = formatChannelType(channel);

        log.info("{} \"{}\" name was updated to {} in guild \"{}\"", type, newValue, oldValue, guild.getName());
        EmbedDescriptionBuilder description = new EmbedDescriptionBuilder("%s Name Updated".formatted(type));
        if (channel.getType() == ChannelType.TEXT || channel.getType() == ChannelType.VOICE) {
            description.appendLine("Channel: %s".formatted(event.getChannel().getAsMention()), true);
        }
        description.appendLine("Name: `%s` -> `%s`".formatted(oldValue, newValue), true);
        logFeature.sendLog(guild, LogType.CHANNEL_CONFIGURATION, EmbedUtils.successEmbed()
                .setDescription(description
                        .build())
                .build());
    }

    @Override
    public void onChannelUpdateNSFW(@NonNull BatGuild guild, @NonNull ChannelUpdateNSFWEvent event) {
        logFeature.sendLog(guild, LogType.CHANNEL_CONFIGURATION, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("Channel NSFW Updated")
                        .appendLine("Channel: %s".formatted(event.getChannel().getAsMention()), true)
                        .appendLine("NSFW: `%s` -> `%s`".formatted(
                                Boolean.TRUE.equals(event.getOldValue()) ? "Yes" : "No",
                                Boolean.TRUE.equals(event.getNewValue()) ? "Yes" : "No"
                        ), true)
                        .build())
                .build());
    }

    @Override
    public void onChannelUpdateUserLimit(@NonNull BatGuild guild, @NonNull ChannelUpdateUserLimitEvent event) {
        if (Objects.equals(event.getOldValue(), event.getNewValue()) || event.getOldValue() == null || event.getNewValue() == null) {
            return;
        }

        VoiceChannel channel = event.getChannel().asVoiceChannel();
        log.info("Voice channel \"{}\" user limit was updated to {} in guild \"{}\"", channel.getName(), event.getNewValue(), guild.getName());
        EmbedDescriptionBuilder description = new EmbedDescriptionBuilder("Voice Channel User Limit Updated")
                .appendLine("Channel: %s".formatted(channel.getAsMention()), true)
                .appendLine("User Limit: `%s` -> `%s`".formatted(
                        event.getOldValue() == 0 ? "Unlimited" : NumberFormatter.format(event.getOldValue()),
                        event.getNewValue() == 0 ? "Unlimited" : NumberFormatter.format(event.getNewValue())
                ), true);
        logFeature.sendLog(guild, LogType.CHANNEL_CONFIGURATION, EmbedUtils.successEmbed()
                .setDescription(description.build())
                .build());
    }

    @Override
    public void onChannelUpdateRegion(@NonNull BatGuild guild, @NonNull ChannelUpdateRegionEvent event) {
        if (Objects.equals(event.getOldValue(), event.getNewValue()) || event.getOldValue() == null || event.getNewValue() == null) {
            return;
        }

        String newRegion = event.getNewValue() == Region.UNKNOWN ? "Automatic" : event.getNewValue().getName();
        log.info("Channel \"{}\" region was updated to \"{}\" in guild \"{}\"", event.getChannel().getName(), newRegion, guild.getName());
        logFeature.sendLog(guild, LogType.CHANNEL_CONFIGURATION, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("Channel Region Updated")
                        .appendLine("Channel: %s".formatted(event.getChannel().getAsMention()), true)
                        .appendLine("Region: `%s` -> `%s`".formatted(
                                event.getOldValue() == Region.UNKNOWN ? "Automatic" : event.getOldValue().getName(),
                                newRegion
                        ), true)
                        .build())
                .build());
    }

    @Override
    public void onChannelUpdateSlowmode(@NonNull BatGuild guild, @NonNull ChannelUpdateSlowmodeEvent event) {
        if (Objects.equals(event.getOldValue(), event.getNewValue()) || event.getOldValue() == null || event.getNewValue() == null) {
            return;
        }
        String oldSlowmode = event.getOldValue() == 0 ? "Disabled" : TimeUtils.format(event.getOldValue() * 1000);
        String newSlowmode = event.getNewValue() == 0 ? "Disabled" : TimeUtils.format(event.getNewValue() * 1000);

        log.info("Text channel \"{}\" slowmode was updated to {} in guild \"{}\"", event.getChannel().getName(), newSlowmode, guild.getName());
        logFeature.sendLog(guild, LogType.CHANNEL_CONFIGURATION, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("Channel Slowmode Updated")
                        .appendLine("Channel: %s".formatted(event.getChannel().getAsMention()), true)
                        .appendLine("Slowmode: `%s` -> `%s`".formatted(oldSlowmode, newSlowmode), true)
                        .build())
                .build());
    }
}
