package cc.fascinated.bat.features.logging.listeners;

import cc.fascinated.bat.common.EmbedDescriptionBuilder;
import cc.fascinated.bat.common.EmbedUtils;
import cc.fascinated.bat.event.EventListener;
import cc.fascinated.bat.features.logging.LogFeature;
import cc.fascinated.bat.features.logging.LogType;
import cc.fascinated.bat.model.BatGuild;
import cc.fascinated.bat.model.BatUser;
import cc.fascinated.bat.service.DiscordService;
import cc.fascinated.bat.service.GuildService;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import net.dv8tion.jda.api.entities.Guild;
import net.dv8tion.jda.api.entities.Role;
import net.dv8tion.jda.api.events.guild.GuildBanEvent;
import net.dv8tion.jda.api.events.guild.GuildUnbanEvent;
import net.dv8tion.jda.api.events.guild.member.GuildMemberJoinEvent;
import net.dv8tion.jda.api.events.guild.member.GuildMemberRemoveEvent;
import net.dv8tion.jda.api.events.guild.member.GuildMemberRoleAddEvent;
import net.dv8tion.jda.api.events.guild.member.GuildMemberRoleRemoveEvent;
import net.dv8tion.jda.api.events.guild.member.update.GuildMemberUpdateBoostTimeEvent;
import net.dv8tion.jda.api.events.guild.member.update.GuildMemberUpdateNicknameEvent;
import net.dv8tion.jda.api.events.guild.member.update.GuildMemberUpdateTimeOutEvent;
import net.dv8tion.jda.api.events.user.update.UserUpdateAvatarEvent;
import net.dv8tion.jda.api.events.user.update.UserUpdateGlobalNameEvent;
import net.dv8tion.jda.api.events.user.update.UserUpdateNameEvent;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Component;

import java.time.OffsetDateTime;
import java.util.List;

/**
 * @author Fascinated (fascinated7)
 */
@Component
@Log4j2(topic = "Logging - Member Listener")
public class MemberListener implements EventListener {
    private final LogFeature logFeature;
    private final GuildService guildService;

    @Autowired
    public MemberListener(@NonNull ApplicationContext context, @NonNull GuildService guildService) {
        this.logFeature = context.getBean(LogFeature.class);
        this.guildService = guildService;
    }

    @Override
    public void onGuildMemberJoin(@NonNull BatGuild guild, @NonNull BatUser user, @NonNull GuildMemberJoinEvent event) {
        if (user.getDiscordUser().isBot())
            log.info("User \"{}\" joined the guild \"{}\"", user.getName(), guild.getDiscordGuild().getName());

        logFeature.sendLog(guild, LogType.MEMBER_JOIN, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("Member Joined")
                        .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                        .appendLine("Username: `%s`".formatted(user.getDiscordUser().getName()), true)
                        .appendLine("Joined Discord: <t:%s:R>".formatted(user.getDiscordUser().getTimeCreated().toEpochSecond()), true)
                        .build())
                .setThumbnail(user.getDiscordUser().getEffectiveAvatarUrl())
                .build());
    }

    @Override
    public void onGuildMemberLeave(@NonNull BatGuild guild, BatUser user, @NonNull GuildMemberRemoveEvent event) {
        if (user == null || user.getDiscordUser().isBot()) return;
        log.info("User \"{}\" left the guild \"{}\"", user.getName(), guild.getDiscordGuild().getName());

        logFeature.sendLog(guild, LogType.MEMBER_LEAVE, EmbedUtils.errorEmbed()
                .setDescription(new EmbedDescriptionBuilder("Member Left")
                        .appendLine("Member: <@%s>".formatted(user.getId()), true)
                        .appendLine("Username: `%s`".formatted(user.getName()), true)
                        .build())
                .build());
    }

    @Override
    public void onGuildMemberUpdateNickname(@NonNull BatGuild guild, @NonNull BatUser user, String oldName, String newName, @NonNull GuildMemberUpdateNicknameEvent event) {
        if (user.getDiscordUser().isBot()) {
            return;
        }
        log.info("User \"{}\" changed their nickname from \"{}\" to \"{}\" in the guild \"{}\"", user.getName(), oldName, newName, guild.getDiscordGuild().getName());

        logFeature.sendLog(guild, LogType.MEMBER_NICKNAME_UPDATE, EmbedUtils.genericEmbed()
                .setDescription(new EmbedDescriptionBuilder("Member Nickname Updated")
                        .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                        .appendLine("Nickname: `%s` -> `%s`".formatted(
                                oldName == null ? user.getName() : oldName,
                                newName == null ? user.getName() : newName
                        ), true)
                        .build())
                .build());
    }

    @Override
    public void onUserUpdateGlobalName(@NonNull BatUser user, String oldName, String newName, @NonNull UserUpdateGlobalNameEvent event) {
        if (user.getDiscordUser().isBot()) {
            return;
        }
        log.info("User \"{}\" changed their global name from \"{}\" to \"{}\"", user.getName(), oldName, newName);

        for (Guild guild : DiscordService.JDA.getGuilds()) {
            BatGuild batGuild = guildService.getGuild(guild.getId());
            if (batGuild == null) continue;
            if (!guild.isMember(user.getDiscordUser())) continue; // User is not in the guild

            logFeature.sendLog(batGuild, LogType.MEMBER_GLOBAL_NAME_UPDATE, EmbedUtils.genericEmbed()
                    .setDescription(new EmbedDescriptionBuilder("Member Name Updated")
                            .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                            .appendLine("Name: `%s` -> `%s`".formatted(
                                    oldName == null ? user.getName() : oldName,
                                    newName == null ? user.getName() : newName
                            ), true)
                            .build())
                    .build());
        }
    }

    @Override
    public void onUserUpdateName(@NonNull BatUser user, String oldName, String newName, @NonNull UserUpdateNameEvent event) {
        if (user.getDiscordUser().isBot()) {
            return;
        }
        log.info("User \"{}\" changed their username from \"{}\" to \"{}\"", user.getName(), oldName, newName);

        for (Guild guild : DiscordService.JDA.getGuilds()) {
            BatGuild batGuild = guildService.getGuild(guild.getId());
            if (batGuild == null) continue;
            if (!guild.isMember(user.getDiscordUser())) continue; // User is not in the guild

            logFeature.sendLog(batGuild, LogType.MEMBER_USERNAME_UPDATE, EmbedUtils.genericEmbed()
                    .setDescription(new EmbedDescriptionBuilder("Member Username Updated")
                            .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                            .appendLine("Username: `%s` -> `%s`".formatted(oldName, newName), true)
                            .build())
                    .build());
        }
    }

    @Override
    public void onUserUpdateAvatar(@NonNull BatUser user, String oldAvatarUrl, String newAvatarUrl, @NonNull UserUpdateAvatarEvent event) {
        if (user.getDiscordUser().isBot()) {
            return;
        }
        log.info("User \"{}\" changed their avatar to \"{}\"", user.getName(), newAvatarUrl);

        for (Guild guild : DiscordService.JDA.getGuilds()) {
            BatGuild batGuild = guildService.getGuild(guild.getId());
            if (batGuild == null) continue;
            if (!guild.isMember(user.getDiscordUser())) continue; // User is not in the guild

            logFeature.sendLog(batGuild, LogType.MEMBER_USERNAME_UPDATE, EmbedUtils.genericEmbed()
                    .setDescription(new EmbedDescriptionBuilder("Member Avatar Updated")
                            .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
//                            .appendLine("Old Avatar: %s".formatted(oldAvatarUrl == null ? "None" : "[avatar](%s)".formatted(oldAvatarUrl)), true)
                            .appendLine("New Avatar: %s".formatted(newAvatarUrl == null ? "Removed" : "[avatar](%s)".formatted(newAvatarUrl)), true)
                            .build())
                    .setThumbnail(newAvatarUrl)
                    .build());
        }
    }

    @Override
    public void onGuildMemberRoleAdd(@NonNull BatGuild guild, @NonNull BatUser user, @NonNull List<Role> rolesAdded, @NonNull GuildMemberRoleAddEvent event) {
        if (user.getDiscordUser().isBot()) {
            return;
        }
        log.info("User \"{}\" was given {} roles in the guild \"{}\"", user.getName(), rolesAdded.size(), guild.getDiscordGuild().getName());

        StringBuilder roles = new StringBuilder();
        for (Role role : rolesAdded) {
            roles.append(role.getAsMention()).append(", ");
        }

        String s = rolesAdded.size() > 1 ? "s" : "";
        logFeature.sendLog(guild, LogType.MEMBER_ROLE_UPDATE, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("Member Role%s Added".formatted(s))
                        .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                        .appendLine("Role%s Added: %s".formatted(s, roles.substring(0, roles.length() - 2)), true)
                        .build())
                .build());
    }

    @Override
    public void onGuildMemberRoleRemove(@NonNull BatGuild guild, @NonNull BatUser user, @NonNull List<Role> rolesAdded, @NonNull GuildMemberRoleRemoveEvent event) {
        if (user.getDiscordUser().isBot()) {
            return;
        }
        log.info("User \"{}\" had {} roles removed in the guild \"{}\"", user.getName(), rolesAdded.size(), guild.getDiscordGuild().getName());

        StringBuilder roles = new StringBuilder();
        for (Role role : rolesAdded) {
            roles.append(role.getAsMention()).append(", ");
        }

        String s = rolesAdded.size() > 1 ? "s" : "";
        logFeature.sendLog(guild, LogType.MEMBER_ROLE_UPDATE, EmbedUtils.errorEmbed()
                .setDescription(new EmbedDescriptionBuilder("Member Role%s Removed".formatted(s))
                        .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                        .appendLine("Role%s Removed: %s".formatted(s, roles.substring(0, roles.length() - 2)), true)
                        .build())
                .build());
    }

    @Override
    public void onGuildMemberBan(@NonNull BatGuild guild, @NonNull BatUser user, @NonNull GuildBanEvent event) {
        if (user.getDiscordUser().isBot()) {
            return;
        }
        log.info("User \"{}\" was banned from the guild \"{}\"", user.getName(), guild.getDiscordGuild().getName());

        logFeature.sendLog(guild, LogType.MEMBER_BAN, EmbedUtils.errorEmbed()
                .setDescription(new EmbedDescriptionBuilder("Member Banned")
                        .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                        .build())
                .build());
    }

    @Override
    public void onGuildMemberUnban(@NonNull BatGuild guild, @NonNull BatUser user, @NonNull GuildUnbanEvent event) {
        if (user.getDiscordUser().isBot()) {
            return;
        }
        log.info("User \"{}\" was unbanned from the guild \"{}\"", user.getName(), guild.getDiscordGuild().getName());

        logFeature.sendLog(guild, LogType.MEMBER_UNBAN, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("Member Unbanned")
                        .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                        .build())
                .build());
    }

    @Override
    public void onGuildMemberTimeout(@NonNull BatGuild guild, @NonNull BatUser user, @NonNull GuildMemberUpdateTimeOutEvent event) {
        OffsetDateTime timeoutEnd = event.getNewTimeOutEnd();
        if (user.getDiscordUser().isBot() || timeoutEnd == null) return;
        log.info("User \"{}\" was timed out until \"{}\"", user.getName(), timeoutEnd);

        long seconds = timeoutEnd.toInstant().getEpochSecond();
        logFeature.sendLog(guild, LogType.MEMBER_TIMEOUT, EmbedUtils.errorEmbed()
                .setDescription(new EmbedDescriptionBuilder("Member Timed Out")
                        .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                        .appendLine("Timeout End: <t:%s>".formatted(seconds), true)
                        .appendLine("Relative End: <t:%s:R>".formatted(seconds), true)
                        .build())
                .build());
    }

    @Override
    public void onGuildMemberBoostAdd(@NonNull BatGuild guild, @NonNull BatUser user, OffsetDateTime oldBoostTime,
                                      @NonNull OffsetDateTime newBoostTime, @NonNull GuildMemberUpdateBoostTimeEvent event) {
        if (user.getDiscordUser().isBot()) {
            return;
        }
        log.info("User \"{}\" boosted the guild \"{}\" until \"{}\"", user.getName(), guild.getDiscordGuild().getName(), newBoostTime);

        logFeature.sendLog(guild, LogType.MEMBER_BOOSTED, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("Member Boosted")
                        .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                        .build())
                .build());
    }

    @Override
    public void onGuildMemberBoostRemove(@NonNull BatGuild guild, @NonNull BatUser user, @NonNull OffsetDateTime oldBoostTime,
                                         OffsetDateTime newBoostTime, @NonNull GuildMemberUpdateBoostTimeEvent event) {
        if (user.getDiscordUser().isBot()) {
            return;
        }
        if (event.getMember().isBoosting()) { // A boost was removed, but the user is still boosting
            return;
        }
        log.info("User \"{}\" boost expired in the guild \"{}\"", user.getName(), guild.getDiscordGuild().getName());

        logFeature.sendLog(guild, LogType.MEMBER_BOOST_EXPIRE, EmbedUtils.errorEmbed()
                .setDescription(new EmbedDescriptionBuilder("Member Boost Expired")
                        .appendLine("Member: %s".formatted(user.getDiscordUser().getAsMention()), true)
                        .build())
                .build());
    }
}
