package cc.fascinated.bat.features.tmdb;

import cc.fascinated.bat.common.EmbedUtils;
import cc.fascinated.bat.common.NumberFormatter;
import cc.fascinated.bat.features.Feature;
import cc.fascinated.bat.features.tmdb.command.TMDBCommand;
import cc.fascinated.bat.service.CommandService;
import cc.fascinated.bat.service.TMDBService;
import info.movito.themoviedbapi.model.core.Movie;
import info.movito.themoviedbapi.model.core.TvSeries;
import lombok.NonNull;
import net.dv8tion.jda.api.EmbedBuilder;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Component;

import java.util.List;

/**
 * @author Nick (okNick)
 */
@Component
public class TMDBFeature extends Feature {
    @Autowired
    public TMDBFeature(@NonNull ApplicationContext context, @NonNull CommandService commandService) {
        super("TMDB", true);

        super.registerCommand(commandService, context.getBean(TMDBCommand.class));
    }

    /**
     * Create an embed for a movie page.
     *
     * @param tmdbService The TMDB service.
     * @param query The query to search for.
     * @param language The language to search in.
     * @param primaryReleaseYear The primary release year to filter by.
     * @param region The region to search in.
     * @param year The year to filter by.
     * @param movie The movie index.
     * @param adult Whether to include adult content.
     * @return The movie page embed.
     */
    public static EmbedBuilder pageMovie(@NonNull TMDBService tmdbService, @NonNull String query, String language, String primaryReleaseYear, String region, String year, int movie, boolean adult) {
        List<Movie> movieList = tmdbService.lookupMovies(query, adult, language, primaryReleaseYear, region, year);
        if (movieList == null || movieList.isEmpty()) {
            return null;
        }

        // Adjust movie index to stay within bounds
         if (movie >= movieList.size()) {
            movie = movieList.size() - 1; // Set to the last movie if index exceeds list size
        }

        return EmbedUtils.genericEmbed()
                .setAuthor(movieList.get(movie).getTitle(), "https://www.themoviedb.org/movie/%s".formatted(movieList.get(movie).getId()))
                .setThumbnail("https://media.themoviedb.org/t/p/w220_and_h330_face%s".formatted(movieList.get(movie).getPosterPath()))
                .setDescription(movieList.get(movie).getOverview())
                .addField("Release Date", movieList.get(movie).getReleaseDate(), true)
                .addField("Rating", NumberFormatter.format(movieList.get(movie).getVoteAverage()) + "/10", true)
                .addField("Language", movieList.get(movie).getOriginalLanguage(), true)
                .setFooter("Page %s of %s".formatted(movie + 1, movieList.size()));
    }

    /**
     * Create an embed for a series page.
     *
     * @param tmdbService The TMDB service.
     * @param query The query to search for.
     * @param language The language to search in.
     * @param firstAirDateYear The first air date year to filter by.
     * @param year The year to filter by.
     * @param series The series index.
     * @param adult Whether to include adult content.
     * @return The series page embed.
     */
    public static EmbedBuilder pageSeries(@NonNull TMDBService tmdbService, @NonNull String query, String language, int firstAirDateYear, int year, int series, boolean adult) {
        List<TvSeries> seriesList = tmdbService.lookupSeries(query, adult, language, firstAirDateYear, year);
        if (seriesList == null || seriesList.isEmpty()) {
            return null;
        }

        // Adjust series index to stay within bounds
        if (series >= seriesList.size()) {
            series = seriesList.size() - 1; // Set to the last series if index exceeds list size
        }

        return EmbedUtils.genericEmbed()
                .setAuthor(seriesList.get(series).getName(), "https://www.themoviedb.org/tv/%s".formatted(seriesList.get(series).getId()))
                .setThumbnail("https://media.themoviedb.org/t/p/w220_and_h330_face%s".formatted(seriesList.get(series).getPosterPath()))
                .setDescription(seriesList.get(series).getOverview())
                .addField("First Air Date", seriesList.get(series).getFirstAirDate(), true)
                .addField("Rating", NumberFormatter.format(seriesList.get(series).getVoteAverage()) + "/10", true)
                .addField("Language", seriesList.get(series).getOriginalLanguage(), true)
                .setFooter("Page %s of %s".formatted(series + 1, seriesList.size()));
    }
}